# checkstyle: noqa
'''
nommy

A python implementation of Rust's nom.
'''

__title__ = 'nommy'
__version__ = '0.3.2'
__all__ = (
    'parser', 'Data', 'repeating', 'repeating_until_null',
    'le_enum', 'be_enum',
    'NommyError', 'NommyUnpackError', 'NommyLShiftError',
    'NommyEnumValueError', 'NommyChompBitsError', 'NommyFieldError',
    'NommyDataError',
    'string', 'pascal_string', 'flag', 'char',
    'le_u', 'be_u',
    'le_u8', 'be_u8', 'le_i8', 'be_i8', 'bool8',
    'le_u16', 'be_u16', 'le_i16', 'be_i16',
    'le_u32', 'be_u32', 'le_i32', 'be_i32',
    'le_u64', 'be_u64', 'le_i64', 'be_i64',
    'le_float16', 'be_float16',
    'le_float32', 'be_float32',
    'le_float64', 'be_float64',
)
__author__ = 'Johan Nestaas <johannestaas@gmail.com'
__license__ = 'GPLv3'
__copyright__ = 'Copyright 2020 Johan Nestaas'

from .parser import (
    parser, Data, repeating, repeating_until_null,
    string, pascal_string, flag, char,
    le_u, be_u,
    le_u8, be_u8, le_i8, be_i8, bool8,
    le_u16, be_u16, le_i16, be_i16,
    le_u32, be_u32, le_i32, be_i32,
    le_u64, be_u64, le_i64, be_i64,
    le_float16, be_float16,
    le_float32, be_float32,
    le_float64, be_float64,
)
from .enum import le_enum, be_enum
from .exceptions import (
    NommyError, NommyUnpackError, NommyLShiftError, NommyChompBitsError,
    NommyEnumValueError, NommyFieldError, NommyDataError,
)


def main():
    pass



from .exceptions import NommyLShiftError, NommyChompBitsError

# For getting a bit at a point, and masking out the rest.
_BIT_MASK = {
    0: (0b10000000, 0b01111111),
    1: (0b01000000, 0b10111111),
    2: (0b00100000, 0b11011111),
    3: (0b00010000, 0b11101111),
    4: (0b00001000, 0b11110111),
    5: (0b00000100, 0b11111011),
    6: (0b00000010, 0b11111101),
    7: (0b00000001, 0b11111110),
}


class Data:

    def __init__(self, _bytes):
        self._bytes = bytearray(_bytes)
        self._chomped_bits = 0

    def _reset_chomped_bits(self):
        # If we go up to 8, we can just shift our data over by a byte.
        if self._chomped_bits == 8:
            self << 8
            # Now we have 0 chomped bits again.
            self._chomped_bits = 0

    def _convert_bitarray_be(self, arr):
        bit_str = ''.join(str(x) for x in arr)
        return int('0b' + bit_str, 2)

    def chomp_to_bitarray(self, size):
        bits = []
        for chomped in range(size):
            if not self._bytes:
                raise NommyChompBitsError(f'no more data: {self._bytes!r}')
            byt = self._bytes[0]
            mask, rem_mask = _BIT_MASK[self._chomped_bits]
            bit = int(bool(byt & mask))
            bits.append(bit)
            self._bytes[0] &= rem_mask
            self._chomped_bits += 1
            self._reset_chomped_bits()
        return bits

    def chomp_to_bytearray(self, size):
        bit_arr = self.chomp_to_bitarray(size)
        byte_arr = bytearray()
        while bit_arr:
            last = bit_arr[-8:]
            bit_arr = bit_arr[:-8]
            byte_arr.append(self._convert_bitarray_be(last))
        return byte_arr[::-1]

    def _extract_sign_from_bytearray(self, arr, size):
        """
        Not used yet. Two's complement not implemented.
        """
        # If it's size 9, then you might have 0x01ff, signed -255.
        # That first byte is 0x01. 9 % 8 == 1.
        # If it's size 10, then you might have 0x03ff.
        # That first byte is 0b0011. 10 % 8 == 2.
        # For size 11, 0x07ff would be 0111.1111.1111, % 8 == 3...
        mod = size % 8
        # _BIT_MASK key, to get which bit it is.
        # If it was the 3rd bit, like size 11 % 3, then the key would be
        # 5, or 8 - mod.
        # If mod is 0, then it's the 8th bit and the key should be 0.
        if mod == 0:
            key = 0
        else:
            key = 8 - mod
        the_bit, mask = _BIT_MASK[key]
        # Get the signbit.
        signbit = bool(arr[0] & the_bit)
        # Mask it out.
        arr[0] &= mask
        return -1 if signbit else 1

    def _twos_complement(self, arr, size):
        raise NotImplementedError()
        sign = self._extract_sign_from_bytearray(arr, size)
        return sign

    def chomp_bits(self, size, endian='le'):
        arr = self.chomp_to_bytearray(size)
        sign = 1
        # TODO add signed logic and `signed` keyword.
        # if signed:
        #     sign = self._twos_complement(arr, size)
        if endian == 'be':
            arr = arr[::-1]
        val = 0
        for i, b in enumerate(arr):
            val += b * 256**i
        return val * sign

    @property
    def bytes(self):
        return bytes(self._bytes)

    def __lshift__(self, size):
        if size % 8 == 0:
            self._bytes = self._bytes[size // 8:]
        else:
            raise NommyLShiftError(f'cant lshift unless multiple of 8: {size}')



from .parser import le_u, be_u
from .exceptions import NommyEnumValueError

# TODO DRY this up.


def le_enum(size):
    def _enum_decorator(enum):
        chomper = le_u(size)

        def _parser(data, **kwargs):
            val = chomper(data)
            return enum.map(val)

        dct = {}
        for key in enum:
            if key.value in dct:
                raise NommyEnumValueError(
                    f'{enum!r} has value {key.value!r}, which isnt unique'
                )
            dct[key.value] = key

        def _map(v):
            return enum._val_dict.get(v)

        enum._val_dict = dct
        enum.map = _map
        enum.parse = _parser
        enum._is_parser = True

        return enum

    return _enum_decorator


def be_enum(size):
    def _enum_decorator(enum):
        chomper = be_u(size)

        def _parser(data, **kwargs):
            val = chomper(data)
            return enum.map(val)

        dct = {}
        for key in enum:
            if key.value in dct:
                raise NommyEnumValueError(
                    f'{enum!r} has value {key.value!r}, which isnt unique'
                )
            dct[key.value] = key

        def _map(v):
            return enum._val_dict.get(v)

        enum._val_dict = dct
        enum.map = _map
        enum.parse = _parser
        enum._is_parser = True

        return enum

    return _enum_decorator



import struct
from functools import partial
from collections import namedtuple

from .exceptions import NommyUnpackError, NommyFieldError, NommyDataError
from .data import Data


def _make_parser(unpack_str, size):

    def _parser(data, **kwargs):
        try:
            val = struct.unpack(unpack_str, data.bytes[:size])[0]
            data << size * 8
        except struct.error as e:
            raise NommyUnpackError(e)
        return val

    return _parser


char = _make_parser('c', 1)
le_u8 = _make_parser('<B', 1)
be_u8 = _make_parser('>B', 1)
le_i8 = _make_parser('<b', 1)
be_i8 = _make_parser('>b', 1)
bool8 = _make_parser('?', 1)
le_u16 = _make_parser('<H', 2)
be_u16 = _make_parser('>H', 2)
le_i16 = _make_parser('<h', 2)
be_i16 = _make_parser('>h', 2)
le_u32 = _make_parser('<I', 4)
be_u32 = _make_parser('>I', 4)
le_i32 = _make_parser('<i', 4)
be_i32 = _make_parser('>i', 4)
le_u64 = _make_parser('<Q', 8)
be_u64 = _make_parser('>Q', 8)
le_i64 = _make_parser('<q', 8)
be_i64 = _make_parser('>q', 8)
le_float16 = _make_parser('<e', 2)
be_float16 = _make_parser('>e', 2)
le_float32 = _make_parser('<f', 4)
be_float32 = _make_parser('>f', 4)
le_float64 = _make_parser('<d', 8)
be_float64 = _make_parser('>d', 8)
_le_u = {8: le_u8, 16: le_u16, 32: le_u32, 64: le_u64}
_le_i = {8: le_i8, 16: le_i16, 32: le_i32, 64: le_i64}
_be_u = {8: be_u8, 16: be_u16, 32: be_u32, 64: be_u64}
_be_i = {8: be_i8, 16: be_i16, 32: be_i32, 64: be_i64}


def string(size):

    if size is None:
        def _parser(data, **kwargs):
            s = ''
            while data.bytes[0] != 0:
                s += chr(data.bytes[0])
                data << 8
            data << 8
            return s
    else:
        def _parser(data, **kwargs):
            val = data.bytes[:size].decode('utf8')
            data << size * 8
            return val
    return _parser


def pascal_string(data, **kwargs):
    ln = data.bytes[0]
    val = data.bytes[1:ln + 1].decode('utf8')
    data << (ln + 1) * 8
    return val


def flag(data, **kwargs):
    return bool(data.chomp_bits(1))


def le_u(size):
    if size in _le_u:
        return _le_u[size]

    def _parser(data, **kwargs):
        return data.chomp_bits(size, endian='le')

    return _parser


def be_u(size):
    if size in _be_u:
        return _be_u[size]

    def _parser(data, **kwargs):
        return data.chomp_bits(size, endian='be')

    return _parser


# TODO add signed logic for le/be functions.
'''
def le_i(size):

    def _parser(data):
        return data.chomp_bits(size, endian='le', signed=True)

    return _parser


def be_i(size):

    def _parser(data):
        return data.chomp_bits(size, endian='be', signed=True)

    return _parser
'''


class repeating:
    """
    This allows you to have a list of a repeating parser, based on the count
    specified by the value in `field_name`.
    """
    _is_parser = True

    def __init__(self, parse_func_or_cls, field):
        self.field = field
        if getattr(parse_func_or_cls, '_is_parser', False):
            self._parse_cls = parse_func_or_cls
            self._parse_func = parse_func_or_cls.parse
        else:
            self._parse_cls = None
            self._parse_func = parse_func_or_cls

    def _get_count_from_values(self, values):
        if '.' in self.field:
            fields = self.field.split('.')
            try:
                count = values[fields[0]]
            except KeyError:
                raise NommyFieldError(
                    f'couldnt get field {fields[0]!r} from {values!r}'
                )
            for fld in fields[1:]:
                if not hasattr(count, fld):
                    raise NommyFieldError(f'{count!r} has no field {fld!r}')
                count = getattr(count, fld)
        else:
            try:
                count = values[self.field]
            except KeyError:
                raise NommyFieldError(
                    f'couldnt get field {self.field!r} from {values!r}'
                )
        if not isinstance(count, int):
            raise NommyFieldError(
                f'couldnt get count from field {self.field!r}: {count!r}'
            )
        return count

    def parse(self, data, values=None, parent=None, **kwargs):
        values = values or {}
        count = self._get_count_from_values(values)
        val = []
        for _ in range(count):
            result = self._parse_func(data, values=values, **kwargs)
            # Would have `rest` in there as well.
            if self._parse_cls:
                result = result[0]
            val.append(result)
        return val


class repeating_until_null:
    """
    This will keep parsing until it hits a null byte.
    """
    _is_parser = True

    def __init__(self, parse_func_or_cls):
        if getattr(parse_func_or_cls, '_is_parser', False):
            self._parse_cls = parse_func_or_cls
            self._parse_func = parse_func_or_cls.parse
        else:
            self._parse_cls = None
            self._parse_func = parse_func_or_cls

    def parse(self, data, values=None, parent=None, **kwargs):
        values = values or {}
        val = []
        while data.bytes[0:1] != b'\0':
            result = self._parse_func(data, values=values, **kwargs)
            # Would have `rest` in there as well.
            if self._parse_cls:
                result = result[0]
            val.append(result)
        data.chomp_bits(8)
        return val


def _parse(cls, _bytes, values=None, parent=None, **kwargs):
    nt_values = {}
    values = values or {}
    if isinstance(_bytes, Data):
        data = _bytes
    elif isinstance(_bytes, (bytes, bytearray)):
        data = Data(_bytes)
    else:
        raise NommyDataError(f'unknown datatype: {_bytes!r}')
    for name, parse_cls_or_func in cls._parsers.items():
        # If it's a class, or a enum.le_enum, etc.
        if getattr(parse_cls_or_func, '_is_parser', False):
            pcls = parse_cls_or_func
            pfunc = pcls.parse
        else:
            pcls = None
            pfunc = parse_cls_or_func
        try:
            val = pfunc(data, values=values, parent=cls, **kwargs)
        except NommyUnpackError as e:
            raise NommyUnpackError(
                f'failed to unpack {name} from bytes {data.bytes!r}: {e}'
            )
        nt_values[name] = val
        values[name] = val
    if parent:
        return cls(**nt_values)
    else:
        return cls(**nt_values), data.bytes


def parser(cls):
    parts = cls.__annotations__
    parsers = {}
    for key, val in parts.items():
        parsers[key] = val
        if getattr(val, '_is_parser', False):
            val._is_subparser = True
    new = namedtuple(
        cls.__name__,
        list(parts.keys()),
    )
    new._is_parser = True
    new._is_subparser = False
    new._parsers = parsers
    new.parse = partial(_parse, new)
    return new



class NommyError(ValueError):
    pass


class NommyUnpackError(NommyError):
    pass


class NommyLShiftError(NommyError):
    pass


class NommyChompBitsError(NommyError):
    pass


class NommyEnumValueError(NommyError):
    pass


class NommyFieldError(NommyError):
    pass


class NommyDataError(NommyError):
    pass



import os
from setuptools import setup

# nommy
# A python byte and bit parser inspired by Rust's nom.


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="nommy",
    version="0.3.2",
    description="A python byte and bit parser inspired by Rust's nom.",
    author="Johan Nestaas",
    author_email="johannestaas@gmail.com",
    license="GPLv3",
    keywords="bytes struct nom",
    url="https://github.com/johannestaas/nommy",
    packages=['nommy'],
    package_dir={'nommy': 'nommy'},
    long_description=read('README.rst'),
    classifiers=[
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        'Development Status :: 3 - Alpha',
        # 'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        'Environment :: Console',
        'Environment :: X11 Applications :: Qt',
        'Environment :: MacOS X',
        'Environment :: Win32 (MS Windows)',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
    ],
    install_requires=[],
    # entry_points={
    #     'console_scripts': [
    #         'nommy=nommy:main',
    #     ],
    # },
    # If you get errors running setup.py install:
    # zip_safe=False,
    #
    # For including non-python files:
    # package_data={
    #     'nommy': ['templates/*.html'],
    # },
    # include_package_data=True,
)
